  module tawny_module
!    "stand alone" version

!    --------------------  select options ----------------------------------
!    check inverse?
     logical                              :: checkinv   = .true. ! .false. 
!    how to: inversion - with or without intermediate steps/files
     character(len=6)                     :: how_invert = 'intmed' ! 'inonly'

!    -------------------- set "standard" file names to be used -------------    
     character(len=15), parameter         :: ginvold =  'GNRMInv_Old.dat',    &
&                                            ll11inv  = 'CholInv_Old.dat',    &
&                                            ginvnew  = 'GNRMInv_New.dat',    & 
&                                            gnrmnew  = 'GNRM_New.dat   ',    &
&                                            mm11inv  = 'CholInv_New.dat' 
!    operational zero
     real(8)                              :: zero = 1.d-6
!    -------------------- end of user modifications ------------------------

     character(len=6)                     :: how_11stor = '      '
     character(len=6)                     :: how_factor = '      '
     real(8), dimension(:), allocatable   :: l11, gg, ww
     real(8), dimension(:,:), allocatable :: l21, hh
     real                                 :: tt1, tt2
     integer                              :: ii1, ii2

  contains

! ===========================
  subroutine tawny( nn, n1 )
! ===========================

  implicit none
  integer, intent(in)                  :: nn, n1
  integer                              :: n2, n11, nn1, mm, irow, icol, mmm, &
&                                         ii, ij, k, kk , jj, ij2
  real(8)                              :: xx, yy, pivot

  call cpu_time ( tt1 ); call system_clock( count = ii1 )

! allocate/read matrices
! ... new GNRM
  nn1 = nn + 1; mmm = get_mm( nn, nn1 )
  allocate( gg(mmm), ww(nn), stat = ii); if( ii /= 0 ) stop 'alloc gg'
  call rdwr_matrix( 'read ', gnrmnew, nn, gg )
! ... old Cholesky
  n11 = n1 + 1; n2 = nn - n1
  if( n2 < 0 ) then
      write(*,*) 'negative no. of new individuals? n, n1, n2', nn, n1, n2
      write(*,*) 'check files - some mix-up likely'; stop 'TAWNY'
  end if
  xx = nn*n1; yy = n11*(3-n11)/2; ij2 = nint(xx) + nint(yy) - 1 

! ----------------------------------------
! Step 1: Complete Cholesky factorisation
! ----------------------------------------

  if( n1 > 0 ) then
      mmm = get_mm( n1, n11 ); allocate( l11(mmm), l21(n11:nn,n1), stat = ii)
      if( ii /= 0 ) stop 'alloc l11'
      call rdwr_matrix( 'read ', ll11inv,  n1, l11 )   ! read L11^{-1}
      ij = n1 
      do icol = 1, n1                                  !  copy G21 to L21
         l21(:,icol) = gg(ij+1:ij+n2); ij = ij + nn - icol
      end do
      do icol = n1, 1, -1                              ! L21 L11^{-T}
         ww(n11:) = 0.d0; ij = icol
         do k = 1, icol  
            ww(n11:) = ww(n11:) + l11(ij) * l21(:,k); ij = ij + n1 - k
         end do
         l21(:,icol) = ww(n11:)
      end do 
      ij = ij2                                         ! G22.1 
      do irow = n11, nn
         ww(:n1) = l21(irow,:)
         do icol = irow, nn
            ij = ij + 1 
            gg(ij) = gg(ij) - dot_product( ww(:n1), l21(icol,:) )
         end do
      end do
      ij = 0  
      do icol = 1, n1                                  ! L21 L11^{-1}
         ww(n11:) = 0.d0 
         do k = icol, n1
            ij = ij + 1; ww(n11:) = ww(n11:) + l11(ij)* l21(:,k)
         end do
         l21(:,icol) = ww(n11:)
      end do
  end if
  call cholesky( nn, ij2, n11, zero, ww, gg ) ! factor G22.1 -> L22

! -----------------------------------
! Step 2: Inverse of Cholesky factor
! -----------------------------------

! invert L22 
  allocate( hh(n11:nn,n11:nn), stat = ii); if( ii /= 0 ) stop 'alloc hh'
  hh = 0.d0; jj = ij2 
  do irow = n11, nn
     jj = jj + 1; kk = jj
     do k = n11, irow
        ww(k) = gg(kk); kk = kk + nn - k
     end do
     pivot = 1.d0 / ww(irow)
     do icol = n11, irow-1
        hh(irow,icol) = - pivot * dot_product( hh(icol:irow-1,icol),         &
&                                              ww(icol:irow-1) )
     end do
     hh(irow,irow) = pivot
  end do

! assemble L^{-1}
  if( how_invert == 'intmed' ) then 
     kk = 0 
     if( n1 > 0 ) then 
         l21 = matmul( hh, l21 )                                  ! L22^{-T} L21
         ij = 0; ii = n11
         do irow = 1, n1         
            ii = ii - 1; gg( kk+1:kk+ii ) = l11(ij+1:ij+ii)       ! L^{11}
            kk = kk + ii; ij = ij + ii 
            gg( kk+1:kk+n2 ) = -l21(:,irow); kk = kk + n2         ! L^{21}
         end do
     end if
     do irow = n11, nn         
        ii = nn1 - irow; gg(kk+1:kk+ii) = hh(irow:nn,irow); kk = kk + ii! L^{22}
     end do
!    ... write out new inverse
     call rdwr_matrix( 'write', mm11inv, nn,  gg )
  end if

! -------------------------
! Step 3: Inverse of matrix
! -------------------------

  if( n1 > 0 ) then
!     ... read 'old' G11
      call rdwr_matrix( 'read ',ginvold, n1, l11 )
      ij = 0; kk = 0
      do irow = 1, n1
         jj = n11 - irow
         ww(irow:n1) = l11(ij+1:ij+jj); ww(n11:) = l21(:,irow); ij = ij + jj 
         do icol = irow, n1                 !        ... adjustment to G11 
            ww(icol) = ww(icol) + dot_product( ww(n11:), l21(:,icol) )
         end do  
         do icol = n11, nn                  !        ... off-diagonal block G21
            ww(icol) = - dot_product( hh(icol:,icol), ww(icol:) )
         end do
         jj = nn1 - irow; gg(kk+1:kk+jj) = ww(irow:); kk = kk + jj
      end do
  end if
! G22 
  ij = ij2
  do irow = n11, nn
     ww(n11:) = hh(:,irow)
     do icol = irow, nn
        ij = ij + 1; gg(ij) = dot_product( ww(icol:), hh(icol:,icol) )
     end do
  end do
! write out matrix
  call rdwr_matrix( 'write', ginvnew, nn, gg )
  call cpu_time( tt2); call system_clock( count= ii2 )

! Check: multiply inverse with matrix
  if( checkinv ) call check_inverse( nn, gg )

  return
  end subroutine tawny

! ====================================================
  subroutine cholesky( nn, ij2, ibegin, zero, ww, gg )
! ====================================================

  implicit none
  integer, intent(in)                            :: nn, ij2, ibegin
  real(8), intent(in)                            :: zero
  real(8), dimension(nn), intent(inout)          :: ww
  real(8), dimension(*), intent(inout)           :: gg
  integer                                        :: irow, icol, ij, k, kk
  real(8)                                        :: tt, xx

  ij = 0; if( ibegin > 1 ) ij = ij2
  do icol = ibegin, nn
     ij = ij + 1; tt = gg(ij)
     if( tt < zero ) then
         write(*,*)'Cholesky factorisation failed at row', icol, ' pivot=', tt
         write(*,*)'Matrix not safely positive definite?'; stop 'tawny'
     end if
     tt = sqrt(tt); gg(ij) = tt
     do irow = icol + 1, nn
        ij = ij + 1; xx =  gg(ij)/ tt; gg(ij) = xx;  ww(irow) = xx
     end do
     xx = (icol+1)*(2-icol)/2; kk = nn*icol + nint(xx) - 1
     do irow = icol + 1, nn
        xx = ww(irow)
        do k = irow, nn
           kk = kk + 1; gg(kk) = gg(kk) - xx * ww(k)
        end do
     end do
  end do

  return
  end subroutine cholesky

! ===================================
  integer function get_mm( nn, nn1 )
! ===================================

  implicit none
  integer, intent(in) :: nn, nn1
  real(8)             :: xx
  integer             :: ibig

  xx = nn * nn1 / 2
  if( xx > huge(ibig) ) then
      write(*,*)'Size of matrix too big for integer*4 addressing'; stop 'tawny'
  end if
  get_mm = nint(xx)

  return
  end function get_mm

! ==================================
  subroutine check_inverse( nn, gg ) 
! ==================================

  implicit none
  integer, intent(in)                            :: nn
  real(8), dimension(*), intent(inout)           :: gg
  real(8), dimension(:,:), allocatable           :: mm1, mm2
  real(8)                                        :: dd, tt, xx
  integer                                        :: i, j, ij, ii, irow, icol

  allocate( mm1(nn,nn), mm2(nn,nn), stat = ii); if( ii /= 0 ) stop 'alloc mm'
  ij = 0
  do i = 1, nn
     do j = i, nn
        ij = ij+1; mm1(j,i) = gg(ij); mm1(i,j) = gg(ij)
     end do
  end do
  call rdwr_matrix( 'read ', gnrmnew, nn, gg )
  ij = 0
  do i = 1, nn
     do j = i,nn
        ij = ij+1; mm2(j,i) = gg(ij); mm2(i,j) = gg(ij)
     end do
  end do
  mm1 = matmul( mm1, mm2 )

  dd = 0.d0; tt = 0.d0
  do irow = 1, nn
     do icol = 1, nn
      xx = mm1(icol,irow)
      if( irow == icol ) then
          dd = dd + xx
          if( abs( xx-1.d0) > 1.d-6 ) print *,'diag',irow,xx
      else 
          tt = tt + xx
          if( abs( xx) > 1.d-6 ) then 
              print *,'offd', irow, icol, xx; pause
          end if
      end if
     end do
  end do
  dd = dd /dble(nn); xx = nn*(nn-1); tt = tt / xx
  print *,'check inverse: average diagonal element =',dd
  print *,'check inverse: average off-diag element =',tt
  if( abs(dd-1.d0) > 1.d-6 .or. abs(tt) > 1.d-6 ) write(*,*) &
&                        'WARNING: Inverse may be incorrect!'
        
  return
  end subroutine check_inverse

  end module tawny_module
! ==========================================================@C KM 10/2012 =====
