!=============================================================================
module tawny_module

!    Version using BLAS and LAPACK routines

!    --------------------  select options ----------------------------------
!    how to: inversion - with or without intermediate steps/files
     character(len=6)                     :: how_invert = 'intmed' ! 'inonly'
!    how to: cholesky factor/inverse 
     character(len=6)                     :: how_factor = 'dpftrf' ! 'dpotrf'
!    how to: full-store or pack G11
!    character(len=6)                     :: how_11stor = 'fulmat' ! 'packed' 
     character(len=6)                     :: how_11stor =  'packed' 
!    check inverse 
     logical                              :: checkinv   = .true.   ! .false. 

!    -------------------- set "standard" file names to be used -------------    
!    standard file names
     character(len=15), parameter         :: ginvold =  'GNRMInv_Old.dat',    &
&                                            ll11inv  = 'CholInv_Old.dat',    &
&                                            ginvnew  = 'GNRMInv_New.dat',    & 
&                                            gnrmnew  = 'GNRM_New.dat   ',    &
&                                            mm11inv  = 'CholInv_New.dat' 
!    -------------------- end of user modifications ------------------------

     real(8), dimension(:), allocatable   :: l11, gg, ww
     real(8), dimension(:,:), allocatable :: l21, hh, qq
     real                                 :: tt1, tt2
     integer                              :: ii1, ii2

contains

! ===========================                                        KM 10/2012
  subroutine tawny( nn, n1 )
! ===========================

  implicit none
  integer, intent(in)                  :: nn, n1
  integer                              :: n2, n11, nn1, mm, irow, icol,        &
&                                         ii, ij, ij1, k, kk , mmm, i, ll
  real(8)                              :: xx, yy
  real(8), external                    :: ddot

  call cpu_time ( tt1 ); call system_clock( count = ii1 )
! allocate & read matrices
! ... new GNRM
  nn1 = nn + 1; mmm = get_mm( nn, nn1 )
  allocate( gg(mmm), stat=ii); if( ii /= 0 ) stop 'alloc gg'
  call rdwr_matrix( 'read ', gnrmnew, nn, gg )

! ... old Cholesky inverse
  n11 = n1 + 1; n2 = nn - n1
  if( n2 < 0 ) then
      write(*,*) 'negative no. of new individuals? n, n1, n2', nn, n1, n2
      write(*,*) 'check files - some mix-up likely'; stop 'TAWNY'
  end if
  xx = nn*n1; yy = n11*(3-n11)/2; mm = nint(xx) + nint(yy)
 
! No 'new' indviduals; straight inversion
  if( n1 == 0 ) then
      select case( how_11stor )
      case( 'packed' )
         call alloc_l11( mmm )
         call dtpttf( 'N', 'L', nn, gg, l11, ii )
         call dpftrf( 'N', 'L', nn, l11, ii )
         if( ii /= 0 ) call lapck_error( 'dpftrf', .true., ii )
         select case( how_invert )
         case( 'inonly' )                                      ! inverse only
            call dpftri( 'N', 'L', nn, l11, ii )
            if( ii /= 0 ) call lapck_error( 'dpftri', .true., ii )
            call dtfttp( 'N', 'L', n2, l11, gg, ii )
         case( 'intmed' )
            call dtftri( 'N', 'L', 'N', n2, l11, ii )            ! L^{-1}
            if( ii /= 0 ) call lapck_error( 'dpftri', .true., ii )
            call dtfttp( 'N', 'L', nn, l11, gg, ii )
            deallocate( l11, stat=ii); if( ii /= 0 ) stop 'dealloc l11'
            call rdwr_matrix( 'write', mm11inv, nn, gg )         ! write L^{-1}
            allocate( hh(nn,nn), stat = ii); if( ii /= 0 ) stop 'alloc hh'
            hh = 0.d0; call dtpttr( 'L', nn, gg, hh, nn, ii )    ! full store
            call dlauum( 'L',nn, hh, nn, ii )
            call dtrttp( 'L', nn, hh, nn, gg, ii)                ! half-store
         end select

      case( 'fulmat' )
         allocate( hh(nn,nn), stat=ii); if( ii /= 0 ) stop 'alloc hh'
         hh = 0.d0
         select case( how_invert )
         case( 'inonly' )                                        ! inverse only
            select case( how_factor )
            case( 'dpotrf' )
               call dtpttr( 'L', nn, gg, hh, nn, ii )            ! full store G
               call dpotrf( 'L', nn, hh, nn, ii)                 ! L
               if( ii /= 0 ) call lapck_error( 'dpotrf', .true., ii ) 
               call dpotri( 'L', nn, hh, nn, ii )                ! G^{-1}
               if( ii /= 0 ) call lapck_error( 'dpotri', .true., ii )
               call dtrttp( 'L', nn, hh, nn, gg, ii)             ! half-store G
            case( 'dpftrf' )
               call alloc_l11( mmm )
               call dtpttf( 'N', 'L', nn, gg, l11, ii )
               call dpftrf( 'N', 'L', nn, l11, ii )
               if( ii /= 0 ) call lapck_error( 'dpftrf', .true., ii )
               call dpftri( 'N', 'L', nn, l11, ii )
               if( ii /= 0 ) call lapck_error( 'dpftri', .true., ii )
               call dtfttp( 'N', 'L', n2, l11, gg, ii )
               deallocate( l11 )
            end select
         case( 'intmed' )
            call dtpttr( 'L', nn, gg, hh, nn, ii )             ! full store G
            call invert_l22( n11, mm, n2, ll )
            call dtrttp( 'L', nn, hh, nn, gg, ii)              ! half-store
            call rdwr_matrix( 'write', mm11inv, nn, gg )       ! write  L^{-1}
            call dlauum( 'L',nn, hh, nn, ii )                  ! G^{-1}
            call dtrttp( 'L', nn, hh, nn, gg, ii)              ! half-store
         end select
     end select

! Updating
  else 
     mmm = get_mm( n1, n11 ); call alloc_l11( mmm )
     call rdwr_matrix( 'read ', ll11inv,  n1, l11 )            ! old L11^{-1}
     kk = 0; ij = 0; ii = n11
     do irow = 1, n1                                           ! G11 <- L11^{-1}
        ii = ii - 1; ij1 = ij + ii
        call dcopy( ii, l11(ij+1:ij1), 1, gg(kk+1:kk+ii), 1 )   
        ij = ij1; kk = kk + ii + n2
     end do

     select case( how_11stor )
     case( 'fulmat' ) 
        allocate( qq(n1,n1), stat=ii);         if( ii /= 0 ) stop 'alloc qq'
        allocate( l21(n11:nn,n1), stat = ii);  if( ii /= 0 ) stop 'alloc l21'
        allocate( hh(n11:nn,n11:nn), stat=ii); if( ii /= 0 ) stop 'alloc hh'
        qq = 0.d0; hh = 0.d0; ij = 0; kk = n11
        do irow = 1, n1                                             ! G11
           kk = kk - 1; ij1 = ij + kk
           call dcopy( kk, gg(ij+1:ij1), 1, qq(irow:,irow), 1)
           ij = ij1; ij1 = ij + n2
           call dcopy( n2, gg(ij+1:ij1), 1, l21(:,irow), 1 )        ! G21
           ij = ij1       
        end do
        call dtpttr( 'L', n2, gg(mm:), hh, n2, ii )                 ! fullst G22
!       Step 1: Adjust 'new' blocks                               
        call dtrmm( 'R', 'L', 'T', 'N', n2, n1, 1.d0, qq, n1, l21, n2 )
        call dsyrk( 'L', 'N', n2, n1, -1.d0, l21, n2, 1.d0, hh, n2 ) ! G22.1
        call dtrmm( 'R', 'L', 'N', 'N', n2, n1, -1.d0, qq, n1, l21, n2 )
!       Step 2: Inverse of Cholesky factor
        call invert_l22( n11, mm, n2, ll ) 
        call dtrmm( 'L', 'L', 'N', 'N', n2, n1, 1.d0, hh, n2, l21, n2 )
        kk = 0; ii = n11                                             ! L22-T L21
        do irow = 1, n1
           ii = ii - 1; kk = kk + ii; ij1 = kk + n2
           call dcopy( n2, l21(:,irow), 1, gg(kk+1:ij1), 1 )       ! L21^{-1}
           kk = ij1  
        end do
        call dtrttp( 'L', n2, hh, n2, gg(mm:), ii)               ! halfst L22^-1
        call rdwr_matrix( 'write', mm11inv, nn, gg )             ! new L^{-1}
!       Step 3: Inverse
        call rdwr_matrix( 'read ', ginvold, n1, l11(:ll) )       ! old G11^{-1}
        call dtpttr( 'L', n1, l11, qq, n1, ii )                  ! full store 
        call dsyrk( 'L', 'T', n1, n2, 1.d0, l21, n2, 1.d0, qq, n1)
        call dtrmm('L', 'L', 'T', 'N', n2, n1, 1.d0, hh, n2, l21, n2)
        ij = 0; kk = n11
        do irow = 1, n1
           kk = kk - 1; ij1 = ij+ kk
           call dcopy( kk, qq(irow:,irow), 1, gg(ij+1:ij1), 1 )
           ij = ij1; ij1 = ij + n2
           call dcopy( n2, l21(:,irow), 1, gg(ij+1:ij1), 1 )
           ij = ij1
        end do

     case( 'packed' )
        allocate( l21(n11:nn,n1), stat = ii); if( ii /= 0 ) stop 'alloc l21'
        allocate( hh(n11:nn,n11:nn), stat=ii); if( ii /= 0 ) stop 'alloc hh'
        allocate( ww(nn), stat = ii ); if( ii /= 0 ) stop 'alloc ww'
!       Step 1: Complete Cholesky factorisation
        ij = n1 
        do icol = 1, n1                                          ! pick out G21 
           call dcopy( n2, gg(ij+1:ij+n2), 1, l21(:,icol), 1 )
           ij = ij + nn - icol
        end do
        call dtpttr( 'L', n2, gg(mm:), hh, n2, ii )              ! pick out G22
        do icol = n1, 1, -1                                      ! L21 L11^{-T}
           ww(n11:) = 0.d0; ij = icol
           do k = 1, icol  
              call daxpy( n2, l11(ij), l21(:,k), 1, ww(n11:), 1 )
              ij = ij + n1 - k
           end do
           call dcopy( n2, ww(n11:), 1, l21(:,icol), 1 )
        end do
        call dsyrk( 'L', 'N', n2, n1, -1.d0, l21, n2, 1.d0, hh, n2 ) ! G22.1
        ij = 0  
        do icol = 1, n1                                           ! L21 L11^{-1}
           ww(n11:) = 0.d0 
           do k = icol, n1
              ij = ij + 1
              call daxpy( n2, l11(ij), l21(:,k), 1, ww(n11:), 1 )
           end do
           call dcopy( n2, ww(n11:), 1, l21(:,icol), 1 )
        end do
!       Step 2: Inverse of Cholesky factor
        call invert_l22( n11, mm, n2, ll )
        call dtrmm('L', 'L', 'N', 'N', n2, n1, 1.d0, hh, n2, l21, n2)
!       Assemble L-inverse
        kk = 0; ii = n11
        do irow = 1, n1
           ii = ii - 1; kk = kk + ii; ij1 = kk + n2
           call dcopy( n2, -l21(:,irow), 1, gg(kk+1:ij1), 1 )         ! L21^{-1}
           kk = ij1
        end do
        call dtrttp( 'L', n2, hh, n2, gg(mm:), ii)         ! half store L22^{-1}
        call rdwr_matrix( 'write', mm11inv, nn, gg )          ! write out L^{-1}
!       Step 3: GNRM inverse - changed: use RFP format
        call rdwr_matrix( 'read ', ginvold, n1, l11(:ll) )      ! 'old' G11^{-1}
        call dtpttf( 'N', 'L', n1, l11, gg, ii )                !  RFP format
        call dsfrk( 'N', 'L', 'T', n1, n2, 1.d0, l21, n2, 1.d0, gg ) ! rank-k up
        call dtfttp( 'N', 'L', n1, gg, l11, ii )                !  RFP to TP
!       ... off-diagonal block G21
        call dtrmm('L', 'L', 'T', 'N', n2, n1, 1.d0, hh, n2, l21, n2)
        ij = 0; kk = 0; ii = n11
        do irow = 1, n1
           ii = ii - 1; ij1 = ij + ii
           call dcopy( ii, l11(ij+1:ij1), 1, gg(kk+1:kk+ii), 1 )! store in G11
           kk = kk + ii; ij = ij1; ij1 = kk + n2
           call dcopy( n2, -l21(:,irow), 1, gg(kk+1:ij1), 1 )
           kk = ij1
        end do
        deallocate( l11, stat = ii ); if( ii /= 0 ) stop 'dealloc l11'
     end select
!    G22 
     call dlauum( 'L',n2, hh, n2, ii )                       ! L22^{-T} L22^{-1}
     call dtrttp( 'L', n2, hh, n2, gg(mm:), ii)              ! half store
  end if

! write out G^{-1}
  call rdwr_matrix( 'write', ginvnew, nn, gg ) 
  
  call cpu_time( tt2 ); call system_clock( count= ii2 )

! Check: multiply inverse with matrix
  if( checkinv ) call check_inverse( nn, gg )

  return
  end subroutine tawny

! ========================================
  subroutine invert_l22( n11, mm, n2, ll )
! ========================================

  implicit none
  integer, intent(in)    :: n11, mm, n2
  integer, intent(inout) :: ll
  integer                :: ii, kk, i
  real(8)                :: xx

  select case( how_factor )
  case( 'dpotrf' )
     call dpotrf( 'L', n2, hh(n11:,n11:), n2, ii)                ! L22
     if( ii /= 0 ) call lapck_error( 'dpotrf', .true., ii )
     call dtrtri( 'L', 'N', n2, hh(n11:,n11:), n2, ii)           ! L22^{-1}
  case( 'dpftrf' )
     xx = n2*(n2+1)/2; kk = nint(xx); ll = ubound(l11,1)
     if( kk > ll ) call alloc_l11( kk )
     call dtrttf( 'N', 'L', n2, hh(n11:,n11:),n2, l11(:kk), ii )  ! full 2 RFP
     call dpftrf( 'N', 'L', n2, l11(:kk), ii )                    ! L22
     if( ii /= 0 ) call lapck_error( 'dpftrf', .true., ii )
     call dtftri( 'N', 'L', 'N', n2, l11(:kk), ii )               ! L22^{-1}
     if( ii /= 0 ) call lapck_error( 'dpftri', .true., ii )
     call dtfttr( 'N', 'L', n2, l11(:kk), hh(n11:,n11:), n2, ii ) ! RFP 2 full 
  end select

  return
  end subroutine invert_l22

! =========================== 
  subroutine alloc_l11( kk )
! =========================== 

  integer, intent(in) :: kk

  if( .not. allocated(l11) ) then
       allocate( l11(kk), stat = ii); if( ii /= 0 ) stop 'alloc l11'
  else if( ubound(l11,1) /= kk ) then
       deallocate(l11)
       allocate( l11(kk), stat = ii); if( ii /= 0 ) stop 'alloc l11'
  end if

  return
  end subroutine alloc_l11

! =======================================
  subroutine lapck_error( rname, fact, ii )
! =======================================

  character(len=6), intent(in) :: rname
  logical, intent(in)          :: fact
  integer, intent(in)          :: ii

  write(*,*)'Routine "',trim(rname),'" returned error code', ii
  if( fact ) write(*,*)'GRM not safely positive definite?'; stop

  end subroutine lapck_error

! ==================================
  subroutine check_inverse( nn, gg ) 
! ==================================

  implicit none
  integer, intent(in)                            :: nn
  real(8), dimension(nn*(nn+1)/2), intent(inout) :: gg
  real(8), dimension(:,:), allocatable           :: mm1, mm2, mm3
  real(8)                                        :: dd, tt, xx
  integer                                        :: i, j, ij, ii, irow, icol, &
&                                                   nn1, mm

  allocate( mm1(nn,nn), mm2(nn,nn), stat = ii); if( ii /= 0 ) stop 'alloc mm'
  ij = 0
  do i = 1, nn
     do j = i, nn
        ij = ij+1; mm1(j,i) = gg(ij); mm1(i,j) = gg(ij)
     end do
  end do
  call rdwr_matrix( 'read ', gnrmnew, nn, gg )
  ij = 0
  do i = 1, nn
     do j = i,nn
        ij = ij+1; mm2(j,i) = gg(ij); mm2(i,j) = gg(ij)
     end do
  end do
! deallocate( gg )

  allocate( mm3(nn,nn), stat = ii ); if( ii /= 0 ) stop 'alloc mm3'
  call dsymm( 'L', 'U', nn, nn, 1.d0, mm1, nn, mm2, nn, 0.d0,  mm3, nn )

  dd = 0.d0; tt = 0.d0
  do irow = 1, nn
     do icol = 1, nn
      xx = mm3(icol,irow)
      if( irow == icol ) then
          dd = dd + xx
          if( abs( xx-1.d0) > 1.d-6 ) print *,'diag',irow,xx
      else 
          tt = tt + xx
      end if
     end do
  end do
  dd = dd /dble(nn); xx = nn*(nn-1); tt = tt / xx
  print *,'check inverse: average diagonal element =',dd
  print *,'check inverse: average off-diag element =',tt
  if( abs(dd-1.d0) > 1.d-6 .or. abs(tt) > 1.d-6 ) write(*,*) &
&                        'WARNING: Inverse may be incorrect!'

  return
  end subroutine check_inverse

! ===================================
  integer function get_mm( nn, nn1 )
! ===================================

  implicit none
  integer, intent(in) :: nn, nn1
  real(8)             :: xx
  integer             :: ibig

  xx = nn * nn1 / 2
  if( xx > huge(ibig) ) then
      write(*,*)'Size of matrix too big for integer*4 addressing'; stop 'TAWNY'
  end if
  get_mm = nint(xx)

  return
  end function get_mm

  end module tawny_module
! ==========================================================@C KM 10/2012 =====
