! "test_tawny"                                                   K.Meyer 2012
! -------------------   choose version  -------------------------------------
! include 'tawny_ind.f90'             !  standalone version
  include 'tawny_las.f90'             !  BLAS/LAPACK version

! -------------------  choose settings for read/write operations ------------
  module rdwr_settings
!    file format: 'binary' or 'format'
     character(len=6)                     :: fileform =  'format' ! 'binary'
!    no. of elements to be read/written: 'all' or 'nnz' (non-zero)
     character(len=3)                     :: elements =  'nnz'    ! 'all'
!    operational zero
     real(8)                              :: zero = 1.d-8
  end module rdwr_settings

! ===================
  program test_tawny
! ===================

  use tawny_module

  call system_clock( count_rate= ii3 )

! get sizes of 'old' and 'new' matrix
print *,gnrmnew
  call get_size( gnrmnew, nn ); if( nn == 0 ) stop 'no GNRM to be inverted'
  call get_size( ll11inv, n1 )

! write info to screen
  write(*,*)'Program "test_tawny" '
  write(*,*)'Matrix sizes found: Total new matrix  =', nn
  write(*,*)'                    "Old" inverse     =', n1
  write(*,*)'                    "New" rows/col.s  =', nn - n1
  write(*,*)'Settings chosen:    how_invert        = "', how_invert,'"'
  if( how_11stor(1:1) /= ' ' ) &
&     write(*,*)'                    how_11stor        = "', how_11stor,'"'
  if( how_factor(1:1) /= ' ' ) &
&     write(*,*)'                    how_factor        = "', how_factor,'"'

! carry out inversion
  call tawny( nn, n1 )

  write(*,'(a,f12.1 )')' "TAWNY" - CPU Time used (sec)           =', tt2 - tt1
  write(*,'(a,f12.1/)')'         - System Time   (sec)           =',           &
&                                                       (ii2 - ii1) / dble(ii3)

  end program test_tawny

! =================================
  subroutine get_size( fname, nrow)
! =================================

  use rdwr_settings
  implicit none
  character(len=15), intent(in)  :: fname
  integer, intent(out)           :: nrow
  integer                        :: ii, irow, icol
  logical                        :: lex

  nrow = 0
  inquire( file = trim(fname), exist = lex ); if( .not. lex ) return

  select case( fileform)
  case( 'format' )
     open(4, file = trim(fname), status ='old', form = 'formatted',          &
&                                                  action = 'read' )
     select case( elements )
     case( 'nnz' )
        do
           read(4,*,iostat=ii) irow, icol; if( ii /= 0 ) exit
           if( irow > nrow ) nrow = irow; if( icol > nrow ) nrow = icol
        end do
     case( 'all' )
        read(4,*,iostat=ii) nrow
     end select

  case( 'binary' )
     open(4, file = trim(fname), status ='old', form = 'unformatted',         &
&                                                    action = 'read' )
     select case( elements )
     case( 'nnz' )
        do
           read(4,iostat=ii) irow, icol; if( ii /= 0 ) exit
           if( irow > nrow ) nrow = irow; if( icol > nrow ) nrow = icol
        end do
     case( 'all' )
        read(4,iostat=ii) nrow
     end select
  end select
  close(4)

  return
  end subroutine get_size

! ================================================
  subroutine rdwr_matrix( gmode, fname, nrow, cc)
! ================================================

  use rdwr_settings
  implicit none
  character(len=5), intent(in)                       :: gmode
  character(len=15), intent(in)                      :: fname
  integer, intent(in)                                :: nrow
  real(8), dimension(nrow*(nrow+1)/2), intent(inout) :: cc
  character(len=5)                                   :: mode2
  character(len=7)                                   :: sstat
  logical                                            :: lopen
  integer                                            :: ii, irow, icol, ij, m2,&
&                                                       mrow, nelem
  real(8)                                            :: xx
  integer, external                                  :: ihmssf

  sstat = 'unknown'; if( gmode == 'read ' ) sstat = 'old    '
  if( fileform == 'format' ) then
      open(4, file = trim(fname), status = sstat, form = 'formatted',     &
&                                                action = trim(gmode) )
  else
      open(4, file = trim(fname), status = sstat, form = 'unformatted',   &
&                                                 action = trim(gmode) )
  end if

  select case( gmode )
  case( 'read ' )
     mode2 = gmode
     select case( elements )
     case( 'all' )
         if( fileform == 'binary' ) then
             read(4, iostat= ii ) mrow, m2
             if( ii /= 0 .or. mrow /= nrow ) return
             read(4,iostat = ii ) cc(:m2)
             read(4,iostat = ii ) cc(m2+1:)
         else
             read(4,*,iostat=ii ) mrow, m2
             if( ii /= 0 .or. mrow /= nrow ) return
             read(4,*,iostat = ii ) cc(:m2)
             read(4,*,iostat = ii ) cc(m2+1:)
         end if
         if( ii /= 0 ) then
             print *,'error: reading matrix all at once'; stop
         end if
         nelem = nrow*(nrow+1)/2
     case( 'nnz' )
        cc = 0.d0; nelem = 0
        do
           if( fileform == 'binary' ) then
               read(4,iostat=ii) irow, icol, xx
           else
               read(4,*,iostat=ii) irow, icol, xx
           end if
           if( ii /= 0 ) exit
           nelem = nelem + 1; cc( ihmssf(irow,icol,nrow) ) = xx
        end do
     end select
  case( 'write' )
     mode2 = 'wrote'
     select case( elements )
     case( 'all' )
         nelem = nrow*(nrow+1)/2
         if( fileform == 'binary' ) then
             write(4) nrow, nelem/2
             write(4, iostat = ii )  cc(:nelem/2)
             write(4, iostat = ii )  cc(nelem/2+1:)
         else
             write(4,*) nrow
             write(4,*,iostat = ii ) cc
         end if
         if( ii /= 0 ) then
             print *,'error: writing matrix all at once'; stop
         end if
     case( 'nnz' )
         ij = 0; nelem = 0; mode2 = 'wrote'
         do irow = 1, nrow
            do icol = irow, nrow 
                ij = ij + 1; if( abs(cc(ij)) < zero ) cycle
                nelem = nelem + 1
                if( fileform == 'binary' ) then
                    write(4) irow, icol, cc(ij)
                else
                    write(4,*) irow, icol, cc(ij)
                end if
            end do
        end do
     end select
  end select

  close(4)
  write(*,'(3a,t19,2a,i12,a,i6,a)')'"',trim(fname),'"','- ',mode2, nelem,     &
&                                  ' coefficients in', nrow, ' rows/columns'
  return
  end subroutine rdwr_matrix

! =================================
  integer function ihmssf( i, j, n)
! =================================

  implicit none
  integer, intent(in) :: i, j, n
  integer             :: i1, j1

  if(i <= j)then
     i1 = i-1; ihmssf = n*i1-i*i1/2+j
  else
     j1 = j-1; ihmssf = n*j1-j*j1/2+i
  end if

  return
  end function ihmssf

